#!/bin/tclsh
#
# TCL CGI script for transmitting multiple files, with names received from
#   an http POST operation
#
# Written by Dave Dykstra <dwd@bell-labs.com> 13 Feb 1997
#
proc errormsg {msg} {
    global argv0
    set me [file tail $argv0]
    puts \
"X-Multiget-Error: $msg
Content-Type: text/html

<HTML>
<HEAD>
<TITLE>
Error from $me
</TITLE>
<BODY>
<H1>
Error from $me:
</H1>
<H3>
<P>
$msg
</H3>"
    exit
}

if {![info exists env(CONTENT_LENGTH)]} {
    errormsg "No CONTENT_LENGTH environment variable"
}
if {[info exists env(QUERY_STRING)]} {
    set prefix [string trimright $env(QUERY_STRING) "/"]
    if {$prefix != ""} {
	set prefix "$prefix/"
    }
} else {
    set prefix ""
}

puts "Content-Type: application/x-multiget\n"

if {[info commands "unsupported0"] == "unsupported0"} {
    rename unsupported0 copychannel
}
foreach n [split [read stdin $env(CONTENT_LENGTH)] "\n"] {
    if {$n == ""} {
	continue
    }
    set f $prefix$n
    switch -regexp "/$f/" {
	{^//} {
	    errormsg "file must be a relative pathname: $f"
	}
	{/\.\./} {
	    errormsg "file may not contain \"..\" components: $f"
	}
    }
    if {[catch {set fd [open $f "r"]} msg] != 0} {
	errormsg "$msg"
    }
    if {[catch {file stat $f statb} msg] != 0} {
	errormsg "$msg"
    }
    if {$statb(type) != "file"} {
	errormsg "$f is not a file"
    }
    puts "Content-Name: $n"
    puts "Content-Length: $statb(size)\n"
    fconfigure $fd -translation binary
    set stdoutConfig [fconfigure stdout -translation]
    fconfigure stdout -translation binary
    copychannel $fd stdout
    fconfigure stdout -translation $stdoutConfig
    close $fd
    puts ""
}
