#
# Main part of NSBD program
# File created by Dave Dykstra <dwd@bell-labs.com>, 18 Nov 1996
#
#
# Copyright (C) 1996-7 by Dave Dykstra and Lucent Technologies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#  
# If those terms are not sufficient for you, contact the author to 
# discuss the possibility of an alternate license.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

proc nsbdCopyright {} {
return \
{Not-So-Bad Distribution   Copyright 1996-7
Open WWW-based Authenticating Software Distribution
Written by Dave Dykstra <dwd@bell-labs.com>
NSBD home page: http://www.bell-labs.com/nsbd
This is free software, and you are welcome to redistribute it under the terms
of the GNU General Public License.  NSBD comes with absolutely no warranty.
Use the "-license" option for details of the license and the non-warranty.
Bell Labs, Innovations for Lucent Technologies}
}

proc usage {{reason ""}} {
    if {$reason != ""} {
	puts stderr "nsbd: $reason"
    }
    puts stderr "[nsbdVersion]\n[nsbdCopyright]\n[nsbdSynopsis]"
    # return magic errorCode, avoids printing error message
    error "" "" "NSBD SILENT"
}

proc nsbdSynopsis {} {
return \
{nsbd {-help | -?} [topic]
nsbd {-version | -V | -license}
nsbd [*] [-unsigned] [*] {file.npd | -}
nsbd [**] -register {URL | file.nsb | -}
nsbd [**] {-update | -poll} {all | {[*] package [...]}}
nsbd [**] -updateOnePackage package
nsbd [**] {-preview | -fetchAll} {[*] {package | URL | file.nsb | -}} [...]
nsbd [*] -updateComments {file.ncf | file.nrd | file.npd | file.nsb}
nsbd {-getUrl | -postUrl} URL
[*] indicates any keyword=value or supplemental '.ncf' or '.nrd' file.
[**] indicates same plus "-ignoreSecurity".}
}

proc nsbdOperation {} {
return \
{NSBD is an open WWW-based authenticating software distribution system.
NSBD adds two basic things to simple web downloads: it authenticates
maintainers of packages with digital signatures so users can trust where
the packages come from, and it supplies a means of checking for updates
to the packages and automatically downloading and installing the updates.
NSBD uses "Pretty Good(Tm) Privacy" (PGP(Tm)) for authentication.  The
single NSBD program handles the operations done by both maintainers of
packages and users of packages.

This is the basic operation: maintainers of packages make their files
available at some http or ftp URL, create a Nsbd Package Description '.npd'
file to describe the files, and run NSBD interactively with that '.npd'
file to indicate that they have approved the files.  That step creates a
Not-So-Bad '.nsb' file which contains the URLs for all files in the
package, MD5 or SHA1 message digests for each file (also known as secure
hashes; they're like checksums, only with much better security because it
is computationally infeasible to discover different file content to match
an existing message digest), and a digital signature of the whole file
created by PGP.  The maintainer then makes the '.nsb' file also available
at a URL with a MIME Content-type of "application/x-nsbd".

When a user clicks on a hyperlink to the URL of a '.nsb' file in his or her
web browser, the '.mailcap' file of the web browser will invoke NSBD with
the '-register' option which pops up a window from which the package can be
registered and installed.  Then, the user runs 'nsbd -poll' periodically
(such as from cron) to check for changes to the '.nsb' file.  If the
signature has changed and is valid, the files whose message digest
information has changed will be loaded.  If all the loaded files match the
message digest information, they will be installed where the user wants
them, and files that have been removed from the package will be removed
from the user's copy.  The user is thus assured that no matter how insecure
the network, only updates from the authorized maintainer will be
automatically installed because only the authorized maintainer controls the
PGP private key.  The user is also assured that if multiple files have
changed they will be installed all together or not at all.  (Detail note:
in order to be valid, a signature has to be newer than an old signature, to
avoid the unlikely case of a spoofer installing an old valid '.nsb' file).}
}

proc nsbdSecurityModel {} {
return \
{The fundamental security model of NSBD is that the user retains control
and only gives away as much authority as necessary to get the update job
done but no more.  Only authenticated updates from authorized maintainers
are accepted, files are only installed into authorized locations on the
user's disk, and only authorized commands are run during installation.
Maintainers may include in their package suggested values for all of the
installation parameters, but it is up to the user to accept them.
Alternatively, users may selectively give away additional privileges to
maintainers to allow them to remotely update most of the installation
parameters, depending on how much they trust the maintainers (see
"maintainerUpdatableKeys").

These are the installation parameters related to security for each package;
see also the descriptions of each keyword in the different NSBD files:
  1. maintainers - the list of PGP user ids of authorized maintainers that
       may install updates.
  2. installTop - the top level directory in which to install files into.
       No complete paths or ".." components are allowed in installed paths,
       so all files will be strictly below this directory. 
  3. validPaths - list of files and directories underneath installTop that
       may be written.
  4. preUpdateCommands/postUpdateCommands - arbitrary commands to execute
       before and after installation of an update.
If the maintainer suggests values for any of these parameters, users of
packages need to realize that the packages may not work properly if they
decide to set the parameters to something else.

Note that if executable programs are distributed and the user executes
those programs after installation is complete, that program will have all
the same access privileges that the user has.  However, if installation is
done under a privileged user id or on a trusted server which don't execute
the programs themselves, the users can at least trust that the programs
won't interfere with each other.  For example, if an untrustworthy
maintainer is trusted to install an infrequently-used program, users can at
assume that other programs installed under the same privileged user id will
not be tainted by the bad program.

Another important part of the security model is that it is designed to be
useful for redistribution.   For example, Bell Labs has a redistribution
system called "Exptools" in which individuals around the company are
permitted to install packages onto the master copy of the distribution
system.  From there, the packages are distributed to a large number of
servers.  The NSBD ".nsb" files, on which the original maintainers have
manually applied digital signatures authorizing all files in their
packages, are intended to be redistributed along with the packages to
verify authenticity on each end server.}
}

proc nsbdGettingStarted {} {
return \
{When you run nsbd the first time, it will create ~/.nsbd/config.ncf and
~/.nsbd/registry.nrd for you and will ask you where your PGP is and if you
want to auto register for NSBD binary updates.  It will also ask you if you
want it to add a line to your $HOME/.mailcap of the form

    application/x-nsbd; $HOME/nsbd/bin/nsbd -register %s

to be used for launching nsbd from a web browser to register for packages.

If you are only a user of packages and not a maintainer, you don't need to
worry about most keywords and you can completely ignore the '.npd' and
'.nsb' files.  You should just peruse the configuration file keywords to
see if some of them seem useful to you; if you have 'pgp', you must specify
a path to it in the configuration file.  Almost all the keywords in the
Registry Database are filled in from the window that pops up when you
register a package with -register, and you should probably try to understand
most of those (there is also pop-up help for those). 

If you are a maintainer of packages, you will need to understand the
keywords in the '.npd' files.  To create a template package description
file with a comment for each keyword, create a new '.npd' file with the
line "nsbPackageDescription: 1" and run "nsbd -updateComments" on the
file.  If you want to make your package available on the web, you should
also look at the lib/nsbd directory for some helpful CGI scripts you may
want to use.

Most people won't need to learn about the '.nsb' files unless they are
trying to evaluate the security of the NSBD protocol.

Note that only the -register option uses a Graphical User Interface; all other
operations are from the command line.  If you do not have a GUI display
available, you can manually register a package by editting registry.nrd and
filling in information from the '.nsb' file (which you can retrieve with
-getUrl) and then using the -update option.  The most important information
from the '.nsb' file are the keywords packageName, nsbUrl, maintainers, and
validPaths (the latter is normally computed from paths).  You will also
need to manually add any needed pgp keys using 'pgp -ka'."}
}

proc nsbdExamples {} {
return \
{Here's a sample package description file for the nsbd binary
package, "nsbd.npd":
  # every .npd file has to begin with the following line
  nsbPackageDescription: 1
  maintainers:
    Dave Dykstra <dwd@bell-labs.com>
      pgpKeyUrl: http://www.bell-labs.com/user/dwd/dwdpgpkey.txt
  executableTypes: sparc-sun-solaris,mips-sgi-irix
  nsbUrl: http://www.bell-labs.com/project/nsbd/%P.nsb
  topUrl: http://www.bell-labs.com/project/nsbd/binaries
  regSubs:
    # delete all .%E suffixes when installing
    \.%E$
  paths:
    # %E expands to every executableType
    bin/nsbd.%E
    lib/nsbd/
    man/man1/nsbd.1

To update the file to include complete comments for each available
keyword:
  nsbd -updateComments nsbd.npd

To create the '.nsb' file from the package description:
  nsbd nsbd.npd

To register for the package via a GUI:
  nsbd -register http://www.bell-labs.com/project/nsbd/%P.nsb

To poll for and install updates for all packages, good to do from
cron nightly or weekly:
  nsbd -poll all

To manually register for the nsbd solaris binary package, add this
to ~/.nsbd/registry.nrd:
  packages:
    nsbd
      nsbUrl:
        http://www.bell-labs.com/nsbd/nsbd.nsb
      maintainers:
        Dave Dykstra <dwd@bell-labs.com>
      executableTypes:
        sparc-sun-solaris
      validPaths:
        bin/nsbd
        lib/nsbd/
        man/man1/nsbd.1

To automatically make and install after an update to nsbd source,
add this to ~/.nsbd/registry.nrd:
  packages:
    nsbdsrc
      postUpdateCommands:
        cd ~/src/nsbd/unix;configure --prefix ~/nsbd;make install #
The  ending '#' is to ignore the parameter of the name of a file containing
a description of the update.}
}

#
# Apparently this is common practice, standard in tcl8.0's http.tcl
#
if {[info commands "unsupported0"] == "unsupported0"} {
    rename unsupported0 copychannel
}

proc nsbd {args} {
    set code [catch {
	eval nsbdDoit $args
    } string]
    nsbdExitCleanup
    if {$code == 1} {
	nsbdErrorHandler $string
    }
    return $code
}

#
# Top-level handler for errors in nsbd program
#
proc nsbdErrorHandler {err} {
    global errorCode errorInfo guiStarted
    if {$errorCode == "NSBD SILENT"} {
	return
    }
    errormsg $err
    if {$guiStarted} {
	set msg "Error: $err"
	set savErrorInfo $errorInfo
	if {[tk_dialog .tkerror "Error in NSBD" $msg error 0 \
						"Exit" "Stack Trace"] == 1} {
	    append msg "\n\nFirst part of stack trace:"
	    set nlines 0
	    foreach line [split $savErrorInfo "\n"] {
		append msg "\n$line"
		if {[incr nlines] > 15} {
		    break
		}
	    }
	    if {[tk_dialog .tkerror "Error in NSBD" $msg error 0 \
				    "Exit" "Full trace to stderr"] == 1} {
		puts stderr "Stack trace:\n$savErrorInfo"
	    }
	}
    } elseif {$errorCode != "NSBD INTERNAL"} {
	set verboseLevel [verboseLevel]
	if {$verboseLevel >= 7} {
	    debugmsg "Stack trace:\n$errorInfo"
	} elseif {$verboseLevel >= 5} {
	    progressmsg "First part of stack trace (for more use verbose=7):"
	    set nlines 0
	    foreach line [split $errorInfo "\n"] {
		progressmsg $line
		if {[incr nlines] > 15} {
		    break
		}
	    }
	}
    }
}

#
# Handler for an error while running tk
#
proc tkerror {err} {
    # make sure guiStarted is set -- there are some race
    #  conditions where it might have not been
    global guiStarted
    set guiStarted 1
    nsbdErrorHandler $err
    nsbdExitCleanup
    exit 1
}

#
# Clean up before doing direct exit
#
proc nsbdExit {{code 0}} {
    nsbdExitCleanup
    exit $code
}

#
# Things to do to clean up.  This is also called directly from
#  Unix signal handlers.  Save errorInfo/errorCode, because this is
#  often called from error handlers that later want to print out stack.
#
proc nsbdExitCleanup {} {
    global errorInfo errorCode
    set savErrorInfo $errorInfo
    set savErrorCode $errorCode
    scratchClean
    set errorInfo $savErrorInfo
    set errorCode $savErrorCode
}

#
# Raise an nsbd-internal error, one which does not cause a stack trace
#
proc nsbderror {message} {
    error $message "" "NSBD INTERNAL"
}

#
# Turn any error in body into an nsbd-internal error
#
proc nsbdcatch {body} {
    set code [catch {uplevel $body} string]
    if {$code == 1} {
	global errorInfo
	return -code $code -errorinfo $errorInfo -errorcode "NSBD INTERNAL" $string
    }
    return -code $code $string
}

#
# Initialize nsbd and process command line arguments.  nsbd may be run
#  multiple times from within tclsh, so re-initialize anything that may
#  be left over from a previous run
#
proc nsbdDoit {args} {
    ensure-tables-loaded
    ensure-options-loaded
    nrdInit

    global cfgContents nrdContents
    catch {unset cfgContents}
    catch {unset nrdContents}
    global cmdKeylist procNsbType procLogLevel ignoreSecurity unsignedNsbfiles
    set cmdKeylist ""
    set procNsbType ""
    set procLogLevel 4
    set ignoreSecurity 0
    set unsignedNsbfiles 0
    global guiStarted messagesWindow scratchTop scratchFiles
    set guiStarted 0
    set messagesWindow ""
    catch {unset scratchTop}
    set scratchFiles ""

    package require Tclmd5
    package require Tclsha1

    global envCfgKeys env
    foreach key $envCfgKeys {
	set upkey [string toupper $key]
	if {[info exists env($upkey)]} {
	    set cfgContents($key) $env($upkey)
	} else {
	    if {[info exists env($key)]} {
		set cfgContents($key) $env($key)
	    }
	}
    }

    # Process command line keyword settings and info-only options
    eval procargs 1 $args

    if {![info exists cfgContents(nsbdpath)]} {
	set cfgContents(nsbdpath) "~/.nsbd"
    }
    set nsbdpath $cfgContents(nsbdpath)

    global cfgFileName
    set cfgFileName [file join $nsbdpath "config.ncf"]
    set regfile [file join $nsbdpath "registry.nrd"]
    if {[file exists $cfgFileName]} {
	quietProcfile $cfgFileName "cfg"
    } else {
	if {![file isdirectory $nsbdpath]} {
	    nsbdcatch {file mkdir $nsbdpath}
	}
	puts "$cfgFileName did not exist, will create it with template comments."
	puts ""
	puts "To operate securely, NSBD requires PGP."
	puts -nonewline \
	     "Please enter path of PGP if you have it: "
	flush stdout
	set ans [gets stdin]
	if {$ans != ""} {
	    set cfgContents(pgp) $ans
	} else {
	    puts "When you get PGP, put path to it in 'pgp' keyword in $cfgFileName"
	}
	withOpen fd $cfgFileName "w" {
	    nsbdGenTemplate $fd cfg cfgContents
	}

	set mailcap "~/.mailcap"
	if {[file exists $mailcap]} {
	    withOpen fd $mailcap "r" {
		while {[gets $fd line] >= 0} {
		    if {[regexp {^application/x-nsbd;} $line]} {
			set mailcap ""
			break
		    }
		}
	    }
	}
	if {$mailcap != ""} {
	    global argv0
	    set nsbd [file join [pwd] $argv0]
	    if {![file executable $nsbd] || ![file isfile $nsbd]} {
		set nsbd [whereExecutable nsbd]
	    }
	    puts ""
	    if {$nsbd != ""} {
		puts "To register a package by clicking on a link in your web browser,"
		puts "  NSBD requires an entry in your $mailcap file."
		puts -nonewline \
		     {Do you want this entry automatically added? [yes] }
		flush stdout
		set ans [gets stdin]
		if {[regexp {[Nn]} $ans]} {
		    puts "To add it manually, add a line like this to $mailcap:"
		    puts "  application/x-nsbd; $nsbd -register %s"
		} else {
		    puts "Adding a line to $mailcap for launching NSBD registrations from web browser."
		    withOpen fd $mailcap "a" {
			puts $fd ""
			puts $fd "# NSBD generated mailcap entry at [currentTime]"
			puts $fd "application/x-nsbd; $nsbd -register %s"
		    }
		}
	    } else {
		puts "Couldn't figure complete path to nsbd, so won't add $mailcap entry."
		puts "Manually add a line like"
		puts "  application/x-nsbd; /path/to/nsbd -register %s"
	    }
	}
    }
    if {![file exists $regfile]} {
	puts ""
	puts "$regfile did not exist, creating it."
	puts ""
	withOpen fd $regfile "w" {
	    nsbdGenTemplate $fd nrd non_existent_variable
	}
	if {$args == ""} {
	    return
	}
    }

    set hackfile [file join $nsbdpath "hacknsbd.tcl"]
    if {[file exists $hackfile]} {
	debugmsg "Sourcing $hackfile, you're on your own."
	source $hackfile
    }

    # Process all the command line arguments
    eval procargs 0 $args
}

#
# little procedure to set the procLogLevel to debug level when processing
#   behind-the-scenes files
#
proc quietProcfile {filename expectedType {typeArg ""}} {
   global procLogLevel
   set savProcLogLevel $procLogLevel
   set procLogLevel 7
   if {($typeArg != "") && ($expectedType == "nrd")} {
       upvar $typeArg $typeArg
   }
   procfile $filename $expectedType $typeArg
   set procLogLevel $savProcLogLevel
}

#
# Process arguments to nsbd.  The first argument is a boolean set to
#   non-zero if only command line keyword settings and info-only options
#   are to be processed.
# Takes variable number of arguments.
#
proc procargs {preprocOnly args} {
    if {$args == ""} {
	if {$preprocOnly} {
	    return
	}
	usage "no parameters supplied"
    }
    global optionTable infoOnlyOptions
    set optionargc 0
    set invokeOption 0
    foreach arg $args {
	if {$optionargc == 1} {
	    $optionfunc $arg
	    set optionargc 0
	    set invokeOption 0
	} elseif {$optionargc >= 3} {
	    # don't process command line keyword settings
	    lappend optionargs $arg
	} elseif {[set equalidx [string first "=" $arg]] > 0} {
	    proccmdkey [string range $arg 0 [expr $equalidx - 1]] \
		    [string range $arg [expr $equalidx + 1] end] $preprocOnly
	} elseif {$optionargc == 2} {
	    # all non-command line keyword settings invoke the optionfunc
	    $optionfunc $arg
	    set invokeOption 0
	} elseif {([string index $arg 0] == "-") && ($arg != "-")} {
	    if {[info exists optionTable($arg)]} {
		if {$preprocOnly && ([lsearch $infoOnlyOptions $arg] < 0)} {
		    set optionfunc "nullfunc"
		} else {
		    set optionfunc "option$arg"
		}
		set optionargc $optionTable($arg)
		if {$optionargc == 0} {
		    $optionfunc
		} else {
		    set invokeOption 1
		    set optionargs ""
		}
	    } else {
		usage "unrecognized option $arg"
	    }
	} else {
	    if {!$preprocOnly} {procfile $arg}
	}
    }
    if {$invokeOption} {
	$optionfunc $optionargs
    }
}

proc nullfunc {args} {
}


#
# procfile accepts a file or a url and knows when urls are valid.
#   A "-" means standard input.
# If there's a second parameter, it should be the fileType of the type
#   of file that is expected, otherwise any type will be processed
#
proc procfile {filename {expectedType ""} {typeArg ""}} {
    global procNsbType
    if {$filename == "-"} {
	set scratchName [scratchAddName "std"]
	set filename "standard input"
	withOpen fd $scratchName "w" {
	    # copy to a scratch file
	    copychannel "stdin" $fd
	}
    } elseif {$expectedType == "nsb"} {
	# allow a url as a source file; copy it to a local file
	if {$procNsbType == "poll"} {
	    set package $typeArg
	    set ans [urlFile $filename [nrdLookup $package lastServerPollTime]]
	    set scratchName [lindex $ans 0]
	    set serverTime [lindex $ans 1]
	    if {$serverTime != ""} {
		nrdUpdate $package set lastServerPollTime $serverTime
	    } elseif {$scratchName != $filename} {
		# was a url, but there's no server time so it must
		#  have not been modified since the lastServerPollTime
		scratchClean $scratchName
		return
	    }
	} else {
	    set scratchName [lindex [urlFile $filename] 0]
	}
    } else {
	set scratchName $filename
    }
    if {$expectedType == "nsb"} {
	# make sure there aren't any lines too long before processing
	# I wish standard gets would support checking for long lines
	#  so wouldn't have to make an extra pass through the file!
	checkLongLines $scratchName $filename
    }
    set fd [nsbdcatch {open $scratchName "r"}]
    alwaysEvalFor $filename {close $fd} {
	foreach {fileType lineno hassig} [nsbdFileType $fd $expectedType] {}
	global fileKeywords
	set firstKey $fileKeywords($fileType)
	if {$fileType == "nup"} {
	    nsbderror "$firstKey files not supported for reading"
	} elseif {($fileType == "nsb") && ($expectedType == "")} {
	    nsbderror "contents begins with \"$firstKey\"; need -register"
	}
	global procLogLevel
	logmsg $procLogLevel "Processing $filename"
	if {$fileType == "nsb"} {
	    procnsbfile $fd $lineno $filename $typeArg $hassig $scratchName
	} else {
	    if ($hassig) {
		nsbderror "PGP signature only expected on '.nsb' type files"
	    }
	    proc${fileType}file $fd $lineno $filename $typeArg
	}
    }
}

#
# Process a '.npd' NSBD Package Description file
#
proc procnpdfile {fd lineno filename {typeArg ""}} {
    global npdContents nsbContents
    catch {unset npdContents}

    # parse the '.npd' file
    nsbdParseContents $fd npd $lineno npdContents

    # merge in applicable configuration and command line keywords
    global cfgnpdKeys cfgContents
    addKeys $cfgnpdKeys cfgContents npdContents npd
    applyCmdkeys npdContents npd

    if {![info exists npdContents(package)] || ($npdContents(package) == "")} {
	# set default package name
	set npdContents(package) [file root [file tail $filename]]
    }

    # construct '.nsb' contents table
    makeNsbContents npdContents nsbContents

    set nsbFilename $nsbContents(package).nsb
    global unsignedNsbfiles
    if {$unsignedNsbfiles} {
	# create an unsigned nsbfile
	nsbdGenFile $nsbFilename nsb nsbContents
	progressmsg "Created an unsigned $nsbFilename"
    } else {
	# first create an unsigned '.nsb' file, then sign it
	set scratchNsbName [scratchAddName "nsb"]
	nsbdGenFile $scratchNsbName nsb nsbContents

	pgpSignFile $scratchNsbName $nsbFilename

	scratchClean "nsb"
    }
}

#
# Process a '.nrd' NSBD Registry Database file
#
proc procnrdfile {fd lineno filename {contentsName ""}} {
    if {$contentsName == ""} {
	# if no name given, write onto the global nrdContents array
	nsbdParseContents $fd nrd $lineno tmpContents
	global nrdContents nrdKeys
	applyKeys $nrdKeys tmpContents nrdContents nrd
    } else {
	# the contents array name comes from 2 levels up, above procfile
	upvar 2 $contentsName contents
	nsbdParseContents $fd nrd $lineno contents
    }
}

#
# Process a configuration file
#
proc proccfgfile {fd lineno filename {typeArg ""}} {
    global cfgContents cfgKeys
    nsbdParseContents $fd cfg $lineno contents
    applyKeys $cfgKeys contents cfgContents cfg
}

#
# Check for too-long lines in a file
#
proc checkLongLines {infile forfile} {
    set fd [nsbdcatch {open $infile "r"}]
    alwaysEvalFor $forfile {close $fd} {
	while {[set str [read $fd 1024]] != ""} {
	    if {[string first "\n" $str] < 0} {
		nsbderror "contains a too long line"
	    }
	}
    }
}

#
# Process a command line keyword setting
# Value may be a list
#
proc proccmdkey {key value {preprocOnly 0}} {
    global cfgContents cmdKeylist
    if {!$preprocOnly} {
	set foundone 0
	foreach fileType {cfg npd nsb} {
	    upvar #0 ${fileType}Keytable keytable
	    if {[info exists keytable($key)]} {
		set foundone 1
		break
	    }
	}
	if {!$foundone} {
	    warnmsg "warning: unknown command line keyword '$key' ignored"
	}
    }
    applyCmdkey $key $value cfgContents cfg
    lappend cmdKeylist $key $value
}

#
# Apply a single command line keyword setting into the contentsname table
#   of type fileType.  A command line setting overrides existing keywords.
#
proc applyCmdkey {key value contentsname fileType} {
    upvar $contentsname contents
    upvar #0 ${fileType}Keytable keytable
    if {[info exists keytable($key)]} {
	if {$value == ""} {
	    if {[info exists contents($key)]} {
		unset contents($key)
	    }
	} else {
	    if {$keytable($key)} {
		# list is expected
		set contents($key) [split $value ","]
	    } else {
		# value is not expected to be a list
		set contents($key) $value
	    }
	}
    }
}

#
# Apply all comand line keywords onto contentsname table of type fileType.
#
proc applyCmdkeys {contentsname fileType} {
    upvar $contentsname contents
    global cmdKeylist
    foreach {key value} $cmdKeylist {
	applyCmdkey $key $value contents $fileType
    }
}

#
# Retrieve a single command line keyword setting
#
proc getCmdkey {keyname} {
    global cmdKeylist
    foreach {key value} $cmdKeylist {
	if {$key == $keyname} {
	    return $value
	}
    }
    return ""
}

#
# Apply the keys in the fromTable onto the toTable.  Keys that already
#  exist are overridden.
#
proc applyKeys {keys fromTablename toTablename keyfileType} {
    upvar $fromTablename fromTable
    upvar $toTablename toTable
    walkKeys $keys fromTable $keyfileType \
	"applyKeysWalkproc toTable [info level]"
}

proc applyKeysWalkproc {toTablename toTablelevel walkType key value} {
    switch $walkType {
	nonListItem {
	    upvar #$toTablelevel $toTablename toTable
	    set toTable($key) $value
	}
	listTable {
	    upvar #$toTablelevel $toTablename toTable
	    # value is fromTablename
	    upvar $value fromTable
	    set toTable($key) $fromTable($key)
	}
    }
}

#
# Add the keys in the fromTable into the toTable.  Keys that already
#  exist are ignored.
#
proc addKeys {keys fromTablename toTablename keyfileType} {
    upvar $fromTablename fromTable
    upvar $toTablename toTable
    walkKeys $keys fromTable $keyfileType \
	"addKeysWalkproc toTable [info level]"
}

proc addKeysWalkproc {toTablename toTablelevel walkType key value} {
    switch $walkType {
	nonListItem {
	    upvar #$toTablelevel $toTablename toTable
	    if {![info exists toTable($key)]} {
		set toTable($key) $value
	    }
	}
	listTable {
	    upvar #$toTablelevel $toTablename toTable
	    if {![info exists toTable($key)]} {
		# value is fromTablename
		upvar $value fromTable
		set toTable($key) $fromTable($key)
	    }
	}
    }
}

#
# Start up the GUI
#
proc startGui {} {
    global guiStarted
    if {$guiStarted} {
	return
    }
    nsbdcatch {startTk}
    set guiStarted 1
    # disable dialog wrapping by putting in a large value for wrapLength
    option add *Dialog.msg.wrapLength 100i
    # set tabs and backtabs in text widgets to switch focus, just like entry
    #  widgets
    bind Text <Tab> [bind Entry <Tab>]
    bind Text <Shift-Tab> [bind Entry <Shift-Tab>]

    wm withdraw .
    makeMessagesWindow
}

#
# stop the GUI.  I haven't figured out any way to restart it once stopped,
#   so it's generally not a good idea to do this.
#
proc stopGui {} {
    global guiStarted
    set guiStarted 0
    deleteMessagesWindow
    destroy .
}

#
# Return the name of a scratch file based on a short name
#
proc scratchName {name} {
    global scratchTop
    if {![info exists scratchTop]} {
	global cfgContents
	if {[info exists cfgContents(tmp)]} {
	    set tmp $cfgContents(tmp)
	} else {
	    set tmp "/tmp"
	}
	set scratchTop [file join $tmp "n[pid]"]
    }
    return $scratchTop$name
}

#
# Add a name to the scratchFiles list
#

proc addToScratchFiles {scratchFile} {
    global scratchFiles
    debugmsg "Adding scratch file $scratchFile"
    luniqueAppend scratchFiles $scratchFile
}

#
# Remove a name from the scratchFiles list
#

proc removeFromScratchFiles {scratchFile} {
    global scratchFiles
    set idx [lsearch -exact $scratchFiles $scratchFile]
    if {$idx >= 0} {
	debugmsg "Removing $scratchFile from the list of scratch files"
	set scratchFiles [lreplace $scratchFiles $idx $idx]
    }
}

#
# Add the scratch name to the list of scratch files and return the full pathname
#
proc scratchAddName {name} {
    set scratchFile [scratchName $name]
    addToScratchFiles $scratchFile
    return $scratchFile
}

#
# Remove scratch file(s)
#
proc scratchClean {{names ""}} {
    global scratchFiles
    if {$names == ""} {
	# scratchFiles may not exist yet if this is being
	#  called as a result of a very early error
	if {[info exists scratchFiles]} {
	    set names $scratchFiles
	}
    }
    foreach name $names {
	set idx [lsearch -exact $scratchFiles $name]
	if {$idx < 0} {
	    set scratchFile [scratchName $name]
	    set idx [lsearch -exact $scratchFiles $scratchFile]
	} else {
	    set scratchFile $name
	}
	debugmsg "Cleaning scratch file $scratchFile"
	# this needs to be -force so it can recursively remove
	#  the temporaryTop directory
	if {[catch {file delete -force $scratchFile} string] > 0} {
	    warnmsg "error deleting $scratchFile: $string"
	}
	if {$idx >= 0} {
	    set scratchFiles [lreplace $scratchFiles $idx $idx]
	}
    }
}

