#
# Tcl-equivalent functions of NSBD C functions, for use when
#   run directly by tclsh
# File created by Dave Dykstra <dwd@bell-labs.com>, 13 Dec 1996
#
# Copyright (C) 1996-7 by Dave Dykstra and Lucent Technologies
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#  
# If those terms are not sufficient for you, contact the author to 
# discuss the possibility of an alternate license.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

#
# start the Tk toolkit
#
proc startTk {} {
    package require Tk
}

#
# make a symbolic link
#
proc symLink {fromFile toFile} {
    exec ln -s $fromFile [expandTildes $toFile]
}

#
# initialize nsbdnestlevel variables
#

proc nsbdnestlevelInit {} {
    upvar #0 nsbdnestlevel_level nestlevel
    upvar #0 nsbdnestlevel_length nestlength
    set nestlevel 0
    catch {unset nestlength}
    set nestlength(0) 0
}

# nsbdnestlevel figures out the nestlevel in a line of an "nsbd" file.
# parameters are
#    1. line - the line to parse
#    2. linelength - the number of bytes in the line
#    3. lineno - the line number that is being parsed
# returns list of two items: the nesting level and the number of characters
#    of whitespace at the beginning of the line
# this is a good candidate for writing in C for speed
# 

proc nsbdnestlevel {line linelength lineno} {
    upvar #0 nsbdnestlevel_level nestlevel
    upvar #0 nsbdnestlevel_length nestlength
    # nlength is nestlength for this line
    set nlength 0
    # whitechars is number of characters of whitespace (tab counts as 1)
    set whitechars 0
    set chr ""
    for {set idx 0} {$idx < $linelength} {incr idx} {
	set chr [string index $line $idx]
	if {$chr == " "} {
	    incr nlength
	    incr whitechars
	} elseif {$chr == "\t"} {
	    incr nlength [expr {8 - ($nlength % 8)}]
	    incr whitechars
	} else {
	    break
	}
    }
    if {($chr == "#") || ($whitechars == $linelength)} {
	# ignore comments and blank lines
	return "$nestlevel $whitechars"
    }
    set curlength $nestlength($nestlevel)
    if {$nlength != $curlength} {
	if {$nlength > $curlength} {
	    incr nestlevel
	    set nestlength($nestlevel) $nlength
	} else {
	    # must be less, find out which if any previous nest level matches
	    unset nestlength($nestlevel)
	    incr nestlevel -1
	    while {($nestlevel > 0) && ($nlength < $nestlength($nestlevel))} {
		unset nestlength($nestlevel)
		incr nestlevel -1
	    }
	    if {$nlength != $nestlength($nestlevel)} {
		nsbderror "invalid indent level on line $lineno"
	    }
	}
    }
    return "$nestlevel $whitechars"
}

